﻿// FetchClimate request definitions for HTML5 application

(function (FC, $, undefined) {

    // Temporal domain
    FC.TemporalDomain = function (years, yearCellMode, days, dayCellMode, hours, hourCellMode) {
        if (!$.isArray(years) || !$.isArray(days) || !$.isArray(hours))
            throw "Years, days and hours must be arrays";
        if (yearCellMode && years.length < 2)
            throw "At least two years must be specified in cell mode";
        if (dayCellMode && days.length < 2)
            throw "At least two days must be specified in cell mode";
        if (hourCellMode && hours.length < 2)
            throw "At least two hours must be specified in cell mode";
        this.years = years;
        this.yearCellMode = yearCellMode;
        this.days = days;
        this.dayCellMode = dayCellMode;
        this.hours = hours;
        this.hourCellMode = hourCellMode;
    };

    FC.TemporalDomain.prototype.isEqual = function (td) {
        if (this.yearCellMode !== td.yearCellMode ||
            this.dayCellMode !== td.dayCellMode ||
            this.hourCellMode !== td.hourCellMode) {
            return false;
        }

        if (!$.isArray(td.years) || !$.isArray(td.days) || !$.isArray(td.hours)) {
            return false;
        }

        if (this.years.length !== td.years.length ||
            this.days.length !== td.days.length ||
            this.hours.length !== td.hours.length) {
            return false;
        }

        for (var i = 0; i < this.years.length; i++) {
            if (this.years[i] !== td.years[i]) {
                return false;
            }
        }

        for (var i = 0; i < this.days.length; i++) {
            if (this.days[i] !== td.days[i]) {
                return false;
            }
        }

        for (var i = 0; i < this.hours.length; i++) {
            if (this.hours[i] !== td.hours[i]) {
                return false;
            }
        }

        return true;
    };

    FC.TemporalDomain.prototype.fillFetchRequest = function (request) {
        request.Domain = request.Domain || {};
        request.Domain.TimeRegion = {
            Years: this.years,
            Days: this.days,
            Hours: this.hours,
            IsIntervalsGridYears: this.yearCellMode,
            IsIntervalsGridDays: this.dayCellMode,
            IsIntervalsGridHours: this.hourCellMode
        };
    };

    FC.TemporalDomain.prototype.yearAxisLength = function () {
        return this.yearCellMode ? (this.years.length - 1) : this.years.length;
    };

    FC.TemporalDomain.prototype.dayAxisLength = function () {
        return this.dayCellMode ? (this.days.length - 1) : this.days.length;
    };

    FC.TemporalDomain.prototype.hourAxisLength = function () {
        return this.hourCellMode ? (this.hours.length - 1) : this.hours.length;
    };

    // ScatteredPoints spatial domain.
    // ScatteredPoints are constructed from two Numeric arrays of similar length or one array of GeoPoints
    FC.ScatteredPoints = function (a, b) {
        this.spatialRegionType = "Points";
        if (!b) {
            if (!$.isArray(a))
                throw "Argument must be an array";
            this.lats = [];
            this.lons = [];
            for (var i = 0; i < a.length; i++) {
                this.lats.push(a[i].lat);
                this.lons.push(a[i].lon);
            }
        } else {
            if (!$.isArray(a) || !$.isArray(b) || a.length != b.length)
                throw "Lats and lons must be arrays of same length";
            this.lats = a;
            this.lons = b;
        }
    };

    FC.ScatteredPoints.prototype.fillFetchRequest = function (request) {
        request.Domain = request.Domain || {};
        request.Domain.SpatialRegionType = "Points";
        request.Domain.Lats = this.lats;
        request.Domain.Lons = this.lons;
        request.Domain.Lats2 = null;
        request.Domain.Lons2 = null;
    };

    // Grid spatial domain base class
    FC.Grid = function (regionType, a, b, c, d, e, f) {
        this.spatialRegionType = regionType;
        if ($.isArray(a) && $.isArray(b) && !c && !d && !e && !f) {
            this.lats = a;
            this.lons = b;
        } else {
            this.lats = [];
            for (var i = 0; i < c; i++)
                this.lats[i] = a + (b - a) * i / (c - 1);
            this.lons = [];
            for (var i = 0; i < f; i++)
                this.lons[i] = d + (e - d) * i / (f - 1);
        }
    };

    FC.Grid.prototype.fillFetchRequest = function (request) {
        request.Domain = request.Domain || {};
        request.Domain.SpatialRegionType = this.spatialRegionType;
        request.Domain.Lats = this.lats;
        request.Domain.Lons = this.lons;
        request.Domain.Lats2 = null;
        request.Domain.Lons2 = null;
    };

    // CellGrid spatial domain
    // CellGrid is constructed either from arrays of lats and lons or
    // from six arguments: latmin, latmax, latcount, lonmin, lonmax, loncount
    FC.CellGrid = function (a, b, c, d, e, f) {
        FC.Grid.call(this, "CellGrid", a, b, c, d, e, f);
    };

    FC.CellGrid.prototype.fillFetchRequest = FC.Grid.prototype.fillFetchRequest;

    // PointGrid spatial domain
    // PointGrid is constructed either from arrays of lats and lons or
    // from six arguments: latmin, latmax, latcount, lonmin, lonmax, loncount
    FC.PointGrid = function (a, b, c, d, e, f) {
        FC.Grid.call(this, "PointGrid", a, b, c, d, e, f);
    };

    FC.PointGrid.prototype.fillFetchRequest = FC.Grid.prototype.fillFetchRequest;

    FC.CellGrid.prototype.fillFetchRequest = function (request) {
        request.Domain = request.Domain || {};
        request.Domain.SpatialRegionType = "CellGrid";
        request.Domain.Lats = this.lats;
        request.Domain.Lons = this.lons;
        request.Domain.Lats2 = null;
        request.Domain.Lons2 = null;
    };

    // ScattetedCells spatial domain
    // ScatteredCells are constructed from four arguments: latmin, lonmin, latmax, lonmax
    FC.ScatteredCells = function (a, b, c, d) {
        if (!$.isArray(latmin) || !$.isArray(latmax) || !$.isArray(lonmin) || !$.isArray(lonmax))
            throw "All arguments must be arrays";
        if (latmin.length != lonmin.length || lonmin.length != lonmax.length || lonmax.length != latmax.length)
            throw "All arguments must be arrays of the same length";
        this.spatialRegionType = "Cells";
        this.lats = latmin;
        this.lats2 = latmax;
        this.lons = lonmin;
        this.lons2 = lonmax;
    };

    FC.ScatteredCells.prototype.fillFetchRequest = function (request) {
        request.Domain = request.Domain || {};
        request.Domain.SpatialRegionType = "Cells";
        request.Domain.Lats = this.lats;
        request.Domain.Lons = this.lons;
        request.Domain.Lats2 = this.lats2;
        request.Domain.Lons2 = this.lons2;
    };

    // Response to a request. Normally, you don't need to construct these objects manually
    FC.Response = function (request, resultUri) {
        this.rq = request;
        this.uri = resultUri;

        Object.defineProperty(this, "request", {
            get: function () { return this.rq; },
            configurable: false
        });

        Object.defineProperty(this, "resultUri", {
            get: function () { return this.uri; },
            configurable: false
        });
    };

    function getRejectReason(request, textStatus, errorThrown) {
        if (errorThrown)
            return errorThrown;
            // As per jQuery docs: http://api.jquery.com/jQuery.ajax/
        else if (!textStatus || textStatus == "timeout" || textStatus == "abort" || textStatus == "error" || textStatus == "parsererror")
            return "timeout (" + request.timeout / 1000 + " sec.)";
        else
            return textStatus;
    }

    function getNowString() {
        var d = new Date();
        return d.toUTCString() + " +" + d.getMilliseconds() + "ms";
    }

    FC.Response.prototype.getSchema = function () {
        var r = new jQuery.Deferred();
        $.ajax({
            url: this.rq.serviceUrl + "/jsproxy/schema?uri=" + encodeURIComponent(this.uri),
            type: "GET",
            dataType: "json",
            timeout: this.rq.timeout,
            success: function (result) {
                r.resolve(result);
            },
            error: function (jqXHR, textStatus, errorThrown) {
                r.reject(getRejectReason(this.rq, textStatus, errorThrown));
            }
        });
        return r.promise();
    };

    function FilterAndBoundFloatData(a) {
        var min = Number.MAX_VALUE;
        var max = -Number.MAX_VALUE;
        if ($.isArray(a)) {
            var a2 = a[0];
            if ($.isArray(a2))
                for (var i = 0; i < a.length; i++) {
                    var d = a[i];
                    FilterAndBoundFloatData(d);
                    if (min > d.max)
                        min = d.max;
                    else if (min > d.min)
                        min = d.min;
                    if (max < d.min)
                        max = d.min;
                    else if (max < d.max)
                        max = d.max;
                }
            else
                for (var i = 0; i < a.length; i++) {
                    var x = parseFloat(a[i]);
                    if (min > x)
                        min = x;
                    if (max < x)
                        max = x;
                    a[i] = x;
                }
        } else {
            a = parseFloat(a);
            if (min > a)
                min = a;
            if (max < a)
                max = a;
        }
        if (min == Number.MAX_VALUE)
            min = max = Number.NaN;
        a.min = min;
        a.max = max;
        return a;
    }

    FC.Response.prototype.getData = function (names) {
        if (!names) {
            if (this.request.dataSources.length === 1) {
                names = ["values", "sd"];
            }
            else {
                names = ["values", "provenance", "sd"];
            }
        }

        var r = new jQuery.Deferred();
        //console.log(getNowString() + ": Requesting " + names + " of " + this.uri);
        var vars = "";
        for (var i = 0; i < names.length; i++) {
            if (vars.length > 0)
                vars = vars + ",";
            vars += names[i];
        }
        $.ajax({
            url: this.rq.serviceUrl + "/jsproxy/data?uri=" + encodeURIComponent(this.uri) + "&variables=" + encodeURIComponent(vars),
            type: "GET",
            dataType: "json",
            timeout: this.timeout,
            success: function (result) {
                //console.log(getNowString() + ": Received data for " + names);
                for (var i = 0; i < names.length; i++)
                    FilterAndBoundFloatData(result[names[i]]);
                r.resolve(result);
            },
            error: function (jqXHR, textStatus, errorThrown) {
                var reason = getRejectReason(this.rq, textStatus, errorThrown);
                console.log(getNowString() + ": Error getting " + names + ": " + reason);
                r.reject(reason);
            }
        });
        return r.promise();
    };

    FC.Response.prototype.getValues = function () {
        return this.getData("values");
    };

    FC.Response.prototype.getUncertainty = function () {
        return this.getData("sd");
    };

    FC.Response.prototype.getProvenance = function () {
        return this.getData("provenance");
    };

    // Request represents single FetchClimate request
    FC.Request = function (options) {

        var status = "new"; // Status is "new" | "pending" | "calculating" | "receiving" | "completed" | "failed"
        var statusData = null; // Status data is null | position in queue (may be NaN) | progress 0..100 | uri of blob | error message
        var _dataSources = {};

        Object.defineProperty(this, "status", {
            get: function () { return status; },
            set: function (value) { status = value; },
            configurable: false
        });

        Object.defineProperty(this, "statusData", {
            set: function (value) { statusData = value; },
            configurable: false
        });

        Object.defineProperty(this, "positionInQueue", {
            get: function () { return (status != "pending") ? NaN : statusData; },
            configurable: false
        });

        Object.defineProperty(this, "percentCompleted", {
            get: function () {
                if (status === "calculating")
                    return statusData;
                else if (status === "receiving" || status === "completed" || status === "failed")
                    return 100;
                else
                    return 0;
            },
            configurable: false
        });

        Object.defineProperty(this, "errorMessage", {
            get: function () { return (status != "failed") ? "" : statusData; },
            configurable: false
        });

        Object.defineProperty(this, "resultUrl", {
            get: function () { return (status != "completed") ? null : statusData; },
            configurable: false
        });

        Object.defineProperty(this, "dataSources", {
            get: function () {
                return _dataSources;
            }
        });

        this.timeout = options.timeout ? options.timeout : 180000; // Default ajax timeout is 3 minutes
        this.pollInterval = options.pollingInterval ? options.pollingInterval : 10000; // Default polling interval is 10 seconds
        this.serviceUrl = options.serviceUrl ? options.serviceUrl : "";
        _dataSources = options.dataSources || [];

        var requestJSON;
        if (options.rawJSON)
            requestJSON = options.rawJSON;
        else {

            this.spatial = options.spatial;
            this.temporal = options.temporal;
            this.variable = options.variable;

            requestJSON = {
                EnvironmentVariableName: this.variable,
                Domain: {
                    Mask: null
                },
                ParticularDataSources: options.dataSources ? options.dataSources : {},
                ReproducibilityTimestamp: options.timestamp ? options.timestamp : Date.UTC(9999, 12, 31, 23, 59, 59, 999)
            };

            this.spatial.fillFetchRequest(requestJSON);
            this.temporal.fillFetchRequest(requestJSON);
        }

        var self = this;

        var performResult = null;
        var statusCallback = null;

        var failRequest = function (message) {
            status = "failed";
            statusData = message;
            if (performResult && performResult.reject) performResult.reject(message);
            performResult = statusCallback = null;
        }

        var onAjaxSuccess = function (answer) {
            // Expected values on stat5: 'pendi' | 'progr' | 'compl' | 'fault'
            var stat5 = answer.substring(0, Math.min(answer.length, 5));

            console.log(getNowString() + ": Status received " + answer);

            if (stat5 === "pendi" || stat5 === "progr") {
                // 'pending', 'progress' responses contain hash.
                var hashIdx = answer.indexOf("hash=");

                if (hashIdx == -1)
                    failRequest("No hash found in response: " + answer);
                else {
                    self.hash = answer.substring(hashIdx + 5).trim();
                }

                if (stat5 === "pendi") {
                    status = "pending";

                    // Get queue position (positive number).
                    statusData = parseInt(answer.match(/pending=(\d+);/)[1], 10);

                    if (statusCallback) {
                        statusCallback(self);
                    }

                    setTimeout(doStatusCheck, self.pollInterval);
                }
                else if (stat5 === "progr") {
                    status = "calculating";

                    // Get progress value.
                    statusData = parseInt(answer.match(/progress=(\d+)%;/)[1], 10);

                    if (statusCallback) {
                        statusCallback(self);
                    }

                    setTimeout(doStatusCheck, self.pollInterval);
                }
            }
            else if (stat5 === "compl") {
                status = "receiving";
                statusData = answer.substring(10);

                var hashIdx = answer.indexOf("Blob=");
                if (hashIdx !== -1) {
                    self.hash = answer.substring(hashIdx + 5).trim();
                }

                if (statusCallback) {
                    statusCallback(self);
                }

                performResult.resolve(new FC.Response(self, statusData));
                performResult = statusCallback = null;
            }
            else if (stat5 === "fault") {
                // 'fault' responses contain hash.
                var hashIdx = answer.indexOf("hash ");

                if (hashIdx == -1)
                    failRequest("No hash found in response: " + answer);
                else {
                    self.hash = answer.substring(hashIdx + 5).trim();
                }

                failRequest(answer);
            }
            else {
                failRequest("Unexpected response: " + answer);
            }
        };

        var onAjaxError = function (jqXHR, textStatus, errorThrown) {
            var reason = getRejectReason(self, textStatus, errorThrown);
            console.log(getNowString() + ": Request failed: " + reason);
            failRequest(reason);
        };

        var doPost = function () {
            console.log(getNowString() + ": Invoking compute for variable " + requestJSON.EnvironmentVariableName);
            status = "pending";
            statusData = NaN; // Position in queue is uknown
            $.ajax({
                url: self.serviceUrl + "/api/compute",
                type: "POST",
                data: JSON.stringify(requestJSON),
                contentType: "application/json; charset=utf-8",
                dataType: "json",
                timeout: self.timeout,
                success: onAjaxSuccess,
                error: onAjaxError
            });
        };

        var doStatusCheck = function () {
            console.log(getNowString() + ": Getting state for " + self.hash);
            $.ajax({
                url: self.serviceUrl + "/api/status?hash=" + self.hash,
                type: "GET",
                dataType: "json",
                timeout: self.timeout,
                success: onAjaxSuccess,
                error: onAjaxError
            });
        };

        this.perform = function (callback) {
            if (!performResult) {
                performResult = new jQuery.Deferred();
                statusCallback = callback;
                doPost();
            }
            return performResult.promise();
        };
    };

    // Appends list of data sources to each variable
    function appendDataSourceInfo(config) {
        for (var i = 0; i < config.EnvironmentalVariables.length; i++) {
            var v = config.EnvironmentalVariables[i];
            v.DataSources = [];
            for (var j = 0; j < config.DataSources.length; j++) {
                var d = config.DataSources[j];
                var found = false;
                for (var k = 0; k < d.ProvidedVariables.length && !found; k++)
                    found = (d.ProvidedVariables[k] == v.Name);
                if (found)
                    v.DataSources.push(d);
            }
        }
        return config;
    }

    function parseCategories(config) {
        var cats = [];
        for (var i = 0; i < config.EnvironmentalVariables.length; i++) {
            var v = config.EnvironmentalVariables[i];
            var p = FC.getCategoriesFromDescription(v.Description);
            v.Description = p.description;
            v.Categories = p.categories;
            for (var j = 0; j < v.Categories.length; j++) {
                var unique = true;
                for (var k = 0; k < cats.length && unique; k++)
                    unique = (cats[k] != v.Categories[j]);
                if (unique)
                    cats.push(v.Categories[j]);
            }
        }
        config.Categories = cats;
        return config;
    }

    FC.getConfigurationTimeout = 180000; // 3 minute timeout for configuration request

    // Returns FetchClimate configuration for given timestamp or latest configuration
    // if timestamp is not specified
    FC.getConfiguration = function (options) {
        options = options || {};
        var r = new jQuery.Deferred();
        $.ajax({
            url: (options.serviceUrl ? options.serviceUrl : "") +
                (options.timestamp ? "/api/configuration?timestamp=" + options.timestamp : "/api/configuration"),
            type: "GET",
            dataType: "json",
            timeout: FC.getConfigurationTimeout,
            success: function (result) {
                r.resolve(parseCategories(appendDataSourceInfo(result)));
            },
            error: function (jqXHR, textStatus, errorThrown) {
                r.reject(getRejectReason(self, textStatus, errorThrown));
            }
        });
        return r.promise();
    };

    FC.getCategoriesFromDescription = function (description) {
        // Work correctly with null
        if (!description)
            return {
                description: "",
                categories: []
            };

        // Split description and categories
        var idx = description.lastIndexOf("Category:");
        var cstr = "";
        if (idx != -1) {
            cstr = description.substring(idx + 9).trim();
            description = description.substring(0, idx).trim();
        } else {
            idx = description.lastIndexOf("Categories:");
            if (idx != -1) {
                cstr = description.substring(idx + 11).trim();
                description = description.substring(0, idx).trim();
            } else
                return {
                    description: description,
                    categories: []
                };
        }
        // Remove trailing ';' if exists
        if (description.charAt(description.length - 1) == ';')
            description = description.substring(0, description.length - 1);

        // Remove trailing '.' if exists.
        if (cstr.charAt(cstr.length - 1) == '.')
            cstr = cstr.substring(0, cstr.length - 1);

        // Split into parts by commas
        var cats = cstr.split(",");
        for (var i = 0; i < cats.length;) {
            cats[i] = cats[i].trim();
            if (cats[i] === "")
                cats.splice(i, 1);
            else
                i++;
        }

        return {
            description: description,
            categories: cats
        };
    };

    FC.tileData = new (function () {
        var that = this;
        var g_latmin = -65, g_latmax = 83.5;
        var g_n = 64; // base tile resolution in intervals
        var g_h = 180.0 / 32 / g_n; // the lowest step at level 5
        var g_view = 48 * g_h;
        this.current_env = "";
        this.current_lat = NaN;
        this.current_lon = NaN;
        this.current_mean = null;
        this.current_upper = null;
        this.current_lower = null;
        this.current_request_count = null;
        // local events implementation
        var events = {};
        this.addEventListener = function (name, handler) {
            if (events.hasOwnProperty(name)) events[name].push(handler);
            else events[name] = [handler];
        };
        var fireEvent = function (name, args) {
            //console.log("%s fire event %s", getNowString(), name);
            if (!events.hasOwnProperty(name)) return;
            if (!args || !args.length) args = [];
            var e = events[name];
            var c = e.length;
            for (var i = 0; i < c; ++i) e[i].apply(null, args);
        };

        this.level0 = $.getJSON("./js/level0.json").then(function (data) {
            function check(a) {
                if (null == a) { return NaN; }
                if ($.isArray(a)) { for (var i = 0; i < a.length; ++i) { a[i] = check(a[i]); } }
                return a;
            };
            for (p in data) { data[p] = check(data[p]); };
            return data;
        }); // a promise to return level0 data
        this.parseTiles = function (latmin, latmax, lonmin, lonmax) {
            // precondition: latmax > latmin
            var n = Math.max(latmax - latmin, Math.abs(lonmax - lonmin)) / g_view;
            var level = n < 1 ? 5 : n < 2 ? 4 : n < 4 ? 3 : n < 8 ? 2 : n < 16 ? 1 : 0;
            var tiles = []; // at level=0 no tiles to indicate
            if (level == 0) { tiles.push([0, 0]); }
            else if (level > 0 && latmax > g_latmin && latmin < g_latmax && lonmin < 180 && lonmax > -180) {
                var scale = Math.pow(2, 5 - level);
                var delta = g_n * g_h * scale;
                var lat1 = Math.floor((latmin - g_latmin) / delta);
                var lat2 = Math.floor((latmax - g_latmin) / delta);
                if (lat1 != lat2) {
                    if (lat1 < 0) lat1 = lat2;
                    if (lat2 * delta >= g_latmax - g_latmin) lat2 = lat1;
                }
                var addlats = function (lon) {
                    tiles.push([lat1, lon]);
                    if (lat2 != lat1) tiles.push([lat2, lon]);
                };
                var lon1 = Math.max(0, Math.floor((lonmin + 180.0) / delta));
                addlats(lon1);
                var lon2 = Math.floor((lonmax + 180.0) / delta);
                if (lon2 * delta + delta > 360) lon2 = lon1;
                if (lon2 != lon1) addlats(lon2);
            }
            return {
                level: level,
                tiles: tiles
            };
            this.getLevel0Tile = function (level, i, j) {
            };
        };
        // a promise to return given tile data
        this.requestTileData = function (env, level, tile) {
            if (level == 0) { return this.level0 }
            else {
                var t12 = new FC.TemporalDomain([1961, 1990], true, FC.monthly, true, [0, 24], true);
                var h = g_h * Math.pow(2, 5 - level)
                var delta = g_n * h;
                var latmin = g_latmin + tile[0] * delta, lonmin = -180 + tile[1] * delta;
                var latmax = latmin + delta;
                var latcount = g_n + 1;
                if (latmax > g_latmax) {
                    latcount = Math.ceil((g_latmax - latmin) / h);
                    latmax = latmin + (latcount - 1) * h;
                }
                // latmin, latmax, latcount, lonmin, lonmax, loncount
                var s = new FC.PointGrid(latmin, latmax, latcount, lonmin, lonmin + delta, g_n + 1);
                var r = new FC.Request({ spatial: s, temporal: t12, variable: env, serviceUrl: "http://fcapp.cloudapp.net" });
                return r.perform()
                    .then(function (response) { return response.getData(["values"]) })
                    .then(function (result) {
                        s[env] = result.values;
                        return s;
                    });
            }
        };
        // data cache contains promises indexed by  env, level and tile. last element is most recently accessed one.
        this.cache = [];
        this.cacheFind = function (env, level, tile) {
            if (level < 1) return null;
            for (var i = 0; i < this.cache.length &&
                !(this.cache[i].env == env &&
                    this.cache[i].level == level &&
                    this.cache[i].tile[0] == tile[0] &&
                    this.cache[i].tile[1] == tile[1]) ; ++i) { }
            if (i < this.cache.length) {
                var found = this.cache[i];
                this.cache.splice(i, 1);
                this.cache.push(found);
                return found.promise;
            }
            else {
                return null;
            }
        };
        this.cacheAdd = function (env, level, tile, promise) {
            //console.log("Cache size: " + String(this.cache.length));
            if (this.cache.length > 200) this.cache.shift();
            this.cache.push({ env: env, level: level, tile: tile, promise: promise });
        }

        this.requestCount = function () {
            var the_request_count = this;
            var count = 0;
            this.beginRequest = function () {
                if (count <= 0) {
                    if (that.current_request_count == the_request_count) fireEvent("beginDownloadData");
                    count = 1;
                }
                else count += 1;
            };
            this.endRequest = function () {
                if (count > 1) count -= 1;
                else {
                    if (that.current_request_count == the_request_count) fireEvent("endDownloadData");
                    count = 0;
                }
            };
        };

        // implementing data cache logic
        this.phoneTileData = function (env, level, tile, lat, lon, id, callback) {
            var current_env = that.current_env;
            var track = that.current_request_count;
            var finalize = function (data) {
                track.endRequest();
                var vv = data[env];
                var lats = data.lats;
                var lons = data.lons;
                var result = { lats: lats, lons: lons, grid: vv, id: id };
                if ((that.current_lat == lat) && (that.current_lon == lon) && (that.current_env == current_env)
                    && (lat >= lats[0]) && (lat <= lats[lats.length - 1])
                    && (lon >= lons[0]) && (lon <= lons[lons.length - 1])) {
                    // get interpolated values
                    // 1. find a grid cell
                    var i = 1; while (i < lats.length - 1 && lats[i] < lat) i += 1;
                    var j = 1; while (j < lons.length - 1 && lons[j] < lon) j += 1;
                    dlat = (lat - lats[i - 1]) / (lats[i] - lats[i - 1]);
                    dlon = (lon - lons[j - 1]) / (lons[j] - lons[j - 1]);
                    var ts = [];
                    for (t = 0; t < 12; t++) {
                        // 2. try bilinear interpolation
                        var tsj = vv[j][i][t] * dlat + vv[j][i - 1][t] * (1 - dlat);
                        var tsjm1 = vv[j - 1][i][t] * dlat + vv[j - 1][i - 1][t] * (1 - dlat);
                        var r = tsj * dlon + tsjm1 * (1 - dlon);
                        if (isNaN(r)) {
                            // 3. collect available points and do inverse point interpolation
                            var d = []; // distance to (dlat, dlon)
                            var z = []; // values
                            if (!isNaN(vv[j - 1][i - 1][t])) { d.push(dlat * dlat + dlon * dlon); z.push(vv[j - 1][i - 1][t]); }
                            if (!isNaN(vv[j][i - 1][t])) { d.push(dlat * dlat + (1 - dlon) * (1 - dlon)); z.push(vv[j][i - 1][t]); }
                            if (!isNaN(vv[j - 1][i][t])) { d.push((1 - dlat) * (1 - dlat) + dlon * dlon); z.push(vv[j - 1][i][t]); }
                            if (!isNaN(vv[j][i][t])) { d.push((1 - dlat) * (1 - dlat) + (1 - dlon) * (1 - dlon)); z.push(vv[j][i][t]); }
                            if (z.length == 1) r = z[0]; // trivial case
                            else if (z.length > 1) {
                                var s=0, w = 0;
                                for (var ii = 0; ii < z.length; ++ii) { d[ii] = 1 / Math.max(0.01, Math.sqrt(d[ii])); w += d[ii]; s += d[ii] * z[ii]; };
                                r = s / w;
                            } // else no data points in the cell
                        }
                        ts[t] = r;
                    }
                    if (id == -1) { that.current_upper = ts; }
                    else if (id == -2) { that.current_lower = ts; }
                    else { that.current_mean = ts; }
                    if ((null != that.current_mean) && (null != that.current_upper) && (null != that.current_lower)) {
                        result.upper = that.current_upper;
                        result.lower = that.current_lower;
                        result.timeseries = that.current_mean;
                    }
                }
                callback(result)
            };

            if (level == 0) {
                track.beginRequest();
                this.level0.then(finalize)
            }
            else {
                var cached = this.cacheFind(env, level, tile);
                if (cached) {
                    track.beginRequest();
                    cached.fail(track.endRequest);
                    cached.then(finalize);
                }
                else {
                    var z = level - 1,
                        zt = [Math.floor(tile[0] / 2), Math.floor(tile[1] / 2)],
                        cz = (z == 0) ? null : this.cacheFind(env, z, zt);
                    while (z > 0 && cz == null) {
                        z -= 1;
                        zt = [Math.floor(zt[0] / 2), Math.floor(zt[1] / 2)];
                        cz = this.cacheFind(env, z, zt);
                    }
                    var promise = (cz ? cz : this.level0);
                    track.beginRequest();
                    promise.fail(track.endRequest);
                    promise.then(function (data) {
                        // extract proper section from an existing tile
                        var scale = Math.pow(2, level - z),
                            count = g_n / scale,
                            lat_off = (tile[0] % scale) * count,
                            lon_off = (z == 0 ? tile[1] : tile[1] % scale) * count,
                            count_lat = Math.min(count, data.lats.length - 1 - lat_off);
                        var lons = [], lats = [], d = data[env], d2 = [];
                        for (var i = 0; i <= count_lat; ++i) lats.push(data.lats[lat_off + i]);
                        for (var i = 0; i <= count; ++i) lons.push(data.lons[lon_off + i]);
                        for (var i = 0; i <= count; ++i) {
                            var r = [];
                            for (var j = 0; j <= count_lat; ++j) r.push(d[lon_off + i][lat_off + j]);
                            d2.push(r);
                        }
                        d = { lats: lats, lons: lons };
                        d[env] = d2;
                        finalize(d);
                    });
                    var promise = this.requestTileData(env, level, tile);
                    track.beginRequest();
                    promise.fail(track.endRequest);
                    promise.then(finalize);
                    this.cacheAdd(env, level, tile, promise);
                }
            }
        };
    })();
    FC.monthly = [1, 32, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366];
    // callback(data)
    FC.phoneData = function (env, lat, latmin, latmax, lon, lonmin, lonmax, callback) {
        FC.tileData.current_env = env;
        FC.tileData.current_lat = lat;
        FC.tileData.current_lon = lon;
        FC.tileData.current_mean = null;
        FC.tileData.current_upper = null;
        FC.tileData.current_lower = null;
        FC.tileData.current_request_count = new FC.tileData.requestCount();
        var tile = FC.tileData.parseTiles(latmin, latmax, lonmin, lonmax);
        for (var t = 0; t < tile.tiles.length; t++) {
            FC.tileData.phoneTileData(env, tile.level, tile.tiles[t], lat, lon, t, callback);
            var limits = FC.Settings.VARIABLE_LIMITS[env];
            if (limits) {
                FC.tileData.phoneTileData(limits[0], tile.level, tile.tiles[t], lat, lon, -1, callback);
                FC.tileData.phoneTileData(limits[1], tile.level, tile.tiles[t], lat, lon, -2, callback);
            }
            else {
                console.log("Error: FC.Settings.VARIABLE_LIMITS[" + env + "] is undefined");
            }
        }
        return tile.tiles.length;
    };
})(window.FC = window.FC || {}, jQuery);
